use rsa::{PublicKey, RSAPublicKey};

use super::*;

impl KeyExt for RSAPublicKey {
    fn to_pem_pkcs8(&self) -> Result<String, KeyError> {
        let bytes = write_pkcs1(self);
        let oid = yasna::models::ObjectIdentifier::from_slice(&RSA_OID);

        let contents = yasna::construct_der(|writer| {
            writer.write_sequence(|writer| {
                writer.next().write_sequence(|writer| {
                    writer.next().write_oid(&oid);
                });
                writer.next().write_bytes(&bytes);
            });
        });

        let p = pem::Pem {
            tag: "PUBLIC KEY".to_owned(),
            contents,
        };

        Ok(pem::encode(&p))
    }

    fn from_pem_pkcs8(pem: &str) -> Result<Self, KeyError> {
        let data = pem::parse(pem).map_err(|_| KeyError::Pem)?;

        if data.tag != "PUBLIC KEY" {
            return Err(KeyError::Kind.into());
        }

        let expected_oid = yasna::models::ObjectIdentifier::from_slice(&RSA_OID);

        let pkey = yasna::parse_der(&data.contents, |reader| {
            reader.read_sequence(|reader| {
                let oid = reader
                    .next()
                    .read_sequence(|reader| reader.next().read_oid())?;

                if oid != expected_oid {
                    return Err(yasna::ASN1Error::new(yasna::ASN1ErrorKind::Invalid));
                }

                let bytes = reader.next().read_bytes()?;

                parse_pkcs1(&bytes)
            })
        })?;

        Ok(pkey)
    }

    fn to_pem_pkcs1(&self) -> Result<String, KeyError> {
        let contents = write_pkcs1(self);

        let p = pem::Pem {
            tag: "RSA PUBLIC KEY".to_owned(),
            contents,
        };

        Ok(pem::encode(&p))
    }

    fn from_pem_pkcs1(pem: &str) -> Result<Self, KeyError> {
        let data = pem::parse(pem).map_err(|_| KeyError::Pem)?;

        if data.tag != "RSA PUBLIC KEY" {
            return Err(KeyError::Kind.into());
        }

        let pkey = parse_pkcs1(&data.contents)?;

        Ok(pkey)
    }
}

fn write_pkcs1(rsa: &RSAPublicKey) -> Vec<u8> {
    yasna::construct_der(|writer| {
        writer.write_sequence(|writer| {
            writer.next().write_biguint(&from_dig(rsa.n()));
            writer.next().write_biguint(&from_dig(rsa.e()));
        })
    })
}

fn parse_pkcs1(bytes: &[u8]) -> Result<RSAPublicKey, yasna::ASN1Error> {
    let (n, e) = yasna::parse_der(bytes, |reader| {
        reader.read_sequence(|reader| {
            let n = reader.next().read_biguint()?;
            let e = reader.next().read_biguint()?;

            Ok((n, e))
        })
    })?;

    RSAPublicKey::new(to_dig(&n), to_dig(&e))
        .map_err(|_| yasna::ASN1Error::new(yasna::ASN1ErrorKind::Invalid))
}
